@@ -15,6 +15,44 @@ from siglab_py.util.simple_math import bucket_series, bucketize_val
 from siglab_py.util.market_data_util import fix_column_types
 from siglab_py.constants import TrendDirection
 
+def classify_candle(
+    candle : pd.Series,
+    min_candle_height_ratio : float = 5,
+    distance_from_mid_doji_threshold_bps : float = 10
+) -> Union[str, None]:
+    candle_class : Union[str, None] = None
+    open = candle['open']
+    high = candle['high']
+    low = candle['low']
+    close = candle['close']
+    candle_full_height = high - low # always positive
+    candle_body_height = close - open # can be negative
+    candle_full_mid = (high + low)/2
+    candle_body_mid = (open + close)/2
+    distance_from_mid_bps = (candle_full_mid/candle_body_mid -1)*10000 if candle_full_mid>candle_body_mid else (candle_body_mid/candle_full_mid -1)*10000
+
+    candle_height_ratio = candle_full_height / abs(candle_body_height) if candle_body_height!=0 else float('inf')
+
+    if (
+        candle_height_ratio>=min_candle_height_ratio
+        and close>low
+    ):
+        candle_class = 'hammer'
+    elif (
+        candle_height_ratio>=min_candle_height_ratio
+        and close<high
+    ):
+        candle_class = 'shooting_star'
+    elif(
+        candle_height_ratio>=min_candle_height_ratio
+        and distance_from_mid_bps<=distance_from_mid_doji_threshold_bps
+    ):
+        candle_class = 'doji'
+
+    # Keep add more ...
+
+    return candle_class
+
 # Fibonacci
 MAGIC_FIB_LEVELS = [0, 0.236, 0.382, 0.5, 0.618, 0.786, 1.00, 1.618, 2.618, 3.618, 4.236]
 
@@ -121,6 +159,12 @@ def compute_candles_stats(
         target_fib_level : float = 0.618,
         pypy_compat : bool = True
         ):
+    BUCKETS_m0_100 = bucket_series(
+						values=list([i for i in range(0,100)]), 
+						outlier_threshold_percent=10, 
+						level_granularity=0.1
+					)
+    
     pd_candles['candle_height'] = pd_candles['high'] - pd_candles['low']
     pd_candles['candle_body_height'] = pd_candles['close'] - pd_candles['open']
 
@@ -139,6 +183,8 @@ def compute_candles_stats(
 
     pd_candles['is_green'] =  pd_candles['close'] >= pd_candles['open']
 
+    pd_candles['candle_class'] = pd_candles.apply(lambda row: classify_candle(row), axis=1) # type: ignore
+
     close_short_periods_rolling = pd_candles['close'].rolling(window=int(sliding_window_how_many_candles/slow_fast_interval_ratio))
     close_long_periods_rolling = pd_candles['close'].rolling(window=sliding_window_how_many_candles)
     close_short_periods_ewm = pd_candles['close'].ewm(span=int(sliding_window_how_many_candles/slow_fast_interval_ratio), adjust=False)
@@ -329,14 +375,14 @@ def compute_candles_stats(
                 first_breach_index = aggressive_mask.idxmax()
                 candle_high = pd_candles.at[first_breach_index, 'high'] 
                 candle_low = pd_candles.at[first_breach_index, 'low']
-                candle_height = candle_high - candle_low
+                candle_height = candle_high - candle_low  # type: ignore
         else:
             aggressive_mask = window['close'] <= window['boillenger_lower_agg']
             if aggressive_mask.any():
                 first_breach_index = aggressive_mask.idxmax()
                 candle_high = pd_candles.at[first_breach_index, 'high'] 
                 candle_low = pd_candles.at[first_breach_index, 'low']
-                candle_height = candle_high - candle_low
+                candle_height = candle_high - candle_low  # type: ignore
 
         return {
             'aggressive_move': aggressive_mask.any(),
@@ -441,7 +487,7 @@ def compute_candles_stats(
             mitigated = pd_candles.iloc[idx + 1:row.name]['close'].lt(row['fvg_high']).any()
         return mitigated
 
-    pd_candles['fvg_mitigated'] = pd_candles.apply(lambda row: compute_fvg_mitigated(row, pd_candles), axis=1)
+    pd_candles['fvg_mitigated'] = pd_candles.apply(lambda row: compute_fvg_mitigated(row, pd_candles), axis=1) # type: ignore
 
     '''
         RSI
@@ -473,13 +519,7 @@ def compute_candles_stats(
         
     lo_rs = lo_ma_up / lo_ma_down
     pd_candles.loc[:,'rsi'] = 100 - (100/(1 + lo_rs))
-    rsi_buckets = bucket_series(
-        values = pd_candles['rsi'].to_list(),
-        outlier_threshold_percent=10,
-        level_granularity=0.1
-    )
-    
-    pd_candles['rsi_bucket'] = pd_candles['rsi'].apply(lambda x: bucketize_val(x, buckets=rsi_buckets))
+    pd_candles['rsi_bucket'] = pd_candles['rsi'].apply(lambda x: bucketize_val(x, buckets=BUCKETS_m0_100))
     pd_candles['ema_rsi'] = pd_candles['rsi'].ewm(
         span=rsi_sliding_window_how_many_candles, 
         adjust=False).mean()
@@ -541,19 +581,43 @@ def compute_candles_stats(
         rsi_sliding_window_how_many_candles if rsi_sliding_window_how_many_candles else sliding_window_how_many_candles).sum()
     pd_candles['money_flow_ratio'] = pd_candles['positive_flow_sum'] / pd_candles['negative_flow_sum']
     pd_candles['mfi'] = 100 - (100 / (1 + pd_candles['money_flow_ratio']))
-    mfi_buckets = bucket_series(
-        values = pd_candles['mfi'].to_list(),
-        outlier_threshold_percent=10,
-        level_granularity=0.1
-    )
-    pd_candles['mfi_bucket'] = pd_candles['mfi'].apply(lambda x: bucketize_val(x, buckets=mfi_buckets))
+    pd_candles['mfi_bucket'] = pd_candles['mfi'].apply(lambda x: bucketize_val(x, buckets=BUCKETS_m0_100))
     
 
     # MACD https://www.investopedia.com/terms/m/macd.asp
     # https://www.youtube.com/watch?v=jmPCL3l08ss
     pd_candles['macd'] = pd_candles['ema_short_periods'] - pd_candles['ema_long_periods']
     pd_candles['signal'] = pd_candles['macd'].ewm(span=int(sliding_window_how_many_candles/slow_fast_interval_ratio), adjust=False).mean()
-    pd_candles['macd_minus_signal'] = pd_candles['macd'] - pd_candles['signal']
+    pd_candles['macd_minus_signal'] = pd_candles['macd'] - pd_candles['signal'] # MACD histogram
+    macd_cur = pd_candles['macd_minus_signal']
+    macd_prev = pd_candles['macd_minus_signal'].shift(1)
+    bullish_macd_crosses = (macd_prev < 0) & (macd_cur > 0)
+    bearish_macd_crosses = (macd_prev > 0) & (macd_cur < 0)
+    pd_candles.loc[bullish_macd_crosses, 'macd_cross'] = 1
+    pd_candles.loc[bearish_macd_crosses, 'macd_cross'] = -1
+    bullish_indices = pd.Series(pd_candles.index.where(pd_candles['macd_cross'] == 1), index=pd_candles.index).astype('Int64')
+    bearish_indices = pd.Series(pd_candles.index.where(pd_candles['macd_cross'] == -1), index=pd_candles.index).astype('Int64')
+    pd_candles['macd_bullish_cross_last_id'] = bullish_indices.rolling(window=pd_candles.shape[0], min_periods=1).max().astype('Int64')
+    pd_candles['macd_bearish_cross_last_id'] = bearish_indices.rolling(window=pd_candles.shape[0], min_periods=1).max().astype('Int64')
+    conditions = [
+        (pd_candles['macd_bullish_cross_last_id'].notna() & 
+        pd_candles['macd_bearish_cross_last_id'].notna() &
+        (pd_candles['macd_bullish_cross_last_id'] > pd_candles['macd_bearish_cross_last_id'])),
+        
+        (pd_candles['macd_bullish_cross_last_id'].notna() & 
+        pd_candles['macd_bearish_cross_last_id'].notna() &
+        (pd_candles['macd_bearish_cross_last_id'] > pd_candles['macd_bullish_cross_last_id'])),
+        
+        (pd_candles['macd_bullish_cross_last_id'].notna() & 
+        pd_candles['macd_bearish_cross_last_id'].isna()),
+        
+        (pd_candles['macd_bearish_cross_last_id'].notna() & 
+        pd_candles['macd_bullish_cross_last_id'].isna())
+    ]
+    choices = ['bullish', 'bearish', 'bullish', 'bearish']
+    pd_candles['macd_cross_last'] = np.select(conditions, choices, default=None) # type: ignore
+    pd_candles.loc[bullish_macd_crosses, 'macd_cross'] = 'bullish'
+    pd_candles.loc[bearish_macd_crosses, 'macd_cross'] = 'bearish'
 
     if not pypy_compat:
         calculate_slope(
