@@ -9,6 +9,7 @@ import math
 import pandas as pd
 import numpy as np
 import asyncio
+from tabulate import tabulate
 
 from ccxt.base.exchange import Exchange as CcxtExchange
 from ccxt import deribit
@@ -83,7 +84,8 @@ async def async_instantiate_exchange(
     secret : str,
     passphrase : str,
     default_type : str,
-    rate_limit_ms : float = 100
+    rate_limit_ms : float = 100,
+    verbose : bool = False
 ) -> Union[AnyExchange, None]:
     exchange : Union[AnyExchange, None] = None
     exchange_name : str = gateway_id.split('_')[0]
@@ -98,7 +100,8 @@ async def async_instantiate_exchange(
                         'rateLimit' : rate_limit_ms,
                         'options' : {
                             'defaultType' : default_type
-                        }
+                        },
+                        'verbose': verbose
                     }
 
     if exchange_name=='binance':
@@ -156,9 +159,10 @@ async def async_instantiate_exchange(
                 "walletAddress" : api_key,
                 "privateKey" : secret,
                 'enableRateLimit'  : True,
-                'rateLimit' : rate_limit_ms
-            }
-        )  # type: ignore
+                'rateLimit' : rate_limit_ms,
+                'verbose': verbose
+            }  # type: ignore
+        )
     else:
         raise ValueError(f"Unsupported exchange {exchange_name}, check gateway_id {gateway_id}.")
 
@@ -192,18 +196,18 @@ def timestamp_to_datetime_cols(pd_candles : pd.DataFrame):
     pd_candles['timestamp_ms'] = pd_candles['timestamp_ms'].apply(_fix_timestamp_ms)
     pd_candles['datetime'] = pd_candles['timestamp_ms'].apply(lambda x: datetime.fromtimestamp(int(x/1000)))
     pd_candles['datetime'] = pd.to_datetime(pd_candles['datetime'])
-    pd_candles['datetime'] = pd_candles['datetime'].dt.tz_localize(None)
+    pd_candles['datetime'] = pd_candles['datetime'].dt.tz_localize(None)  # type: ignore
     pd_candles['datetime_utc'] = pd_candles['timestamp_ms'].apply(
         lambda x: datetime.fromtimestamp(int(x.timestamp()) if isinstance(x, pd.Timestamp) else int(x / 1000), tz=timezone.utc)
     )
     
     # This is to make it easy to do grouping with Excel pivot table
-    pd_candles['year'] = pd_candles['datetime'].dt.year
-    pd_candles['month'] = pd_candles['datetime'].dt.month
-    pd_candles['day'] = pd_candles['datetime'].dt.day
-    pd_candles['hour'] = pd_candles['datetime'].dt.hour
-    pd_candles['minute'] = pd_candles['datetime'].dt.minute
-    pd_candles['dayofweek'] = pd_candles['datetime'].dt.dayofweek  # dayofweek: Monday is 0 and Sunday is 6
+    pd_candles['year'] = pd_candles['datetime'].dt.year  # type: ignore
+    pd_candles['month'] = pd_candles['datetime'].dt.month  # type: ignore
+    pd_candles['day'] = pd_candles['datetime'].dt.day  # type: ignore
+    pd_candles['hour'] = pd_candles['datetime'].dt.hour  # type: ignore
+    pd_candles['minute'] = pd_candles['datetime'].dt.minute  # type: ignore
+    pd_candles['dayofweek'] = pd_candles['datetime'].dt.dayofweek  # type: ignore dayofweek: Monday is 0 and Sunday is 6
 
     pd_candles['week_of_month'] = pd_candles['timestamp_ms'].apply(
         lambda x: timestamp_to_week_of_month(int(x/1000))
@@ -220,8 +224,13 @@ def timestamp_to_datetime_cols(pd_candles : pd.DataFrame):
     )
 
     pd_candles['timestamp_ms_gap'] = pd_candles['timestamp_ms'] - pd_candles['timestamp_ms'].shift(1)
-    timestamp_ms_gap = pd_candles.iloc[-1]['timestamp_ms_gap']
-    assert(pd_candles[~pd_candles.timestamp_ms_gap.isna()][pd_candles.timestamp_ms_gap!=timestamp_ms_gap].shape[0]==0)
+    
+    # Depending on asset, minutes bar may have gaps
+    timestamp_ms_gap_median = pd_candles['timestamp_ms_gap'].median()
+    NUM_MS_IN_1HR = 60*60*1000
+    if timestamp_ms_gap_median>=NUM_MS_IN_1HR:
+        num_rows_with_expected_gap = pd_candles[~pd_candles.timestamp_ms_gap.isna()][pd_candles.timestamp_ms_gap==timestamp_ms_gap_median].shape[0]
+        assert(num_rows_with_expected_gap/pd_candles.shape[0]>0.9)
     pd_candles.drop(columns=['timestamp_ms_gap'], inplace=True)
 
 def timestamp_to_active_trading_regions(
@@ -343,7 +352,7 @@ class NASDAQExchange:
             pd_daily_candles['low'] = pd_daily_candles['low'].astype(str).str.replace('$','')
             pd_daily_candles['close'] = pd_daily_candles['close'].astype(str).str.replace('$','')
             pd_daily_candles['datetime']= pd.to_datetime(pd_daily_candles['datetime'])
-            pd_daily_candles['timestamp_ms'] = pd_daily_candles.datetime.values.astype(np.int64) // 10 ** 6
+            pd_daily_candles['timestamp_ms'] = pd_daily_candles.datetime.values.astype(np.int64) // 10 ** 6  # type: ignore
             pd_daily_candles['symbol'] = symbol
             pd_daily_candles['exchange'] = 'nasdaq'
             fix_column_types(pd_daily_candles)
@@ -362,7 +371,7 @@ class NASDAQExchange:
                 )
 
                 # When you fill foward, a few candles before start date can have null values (open, high, low, close, volume ...)
-                first_candle_dt = pd_hourly_candles[(~pd_hourly_candles.close.isna())  & (pd_hourly_candles['datetime'].dt.time == pd.Timestamp('00:00:00').time())].iloc[0]['datetime']
+                first_candle_dt = pd_hourly_candles[(~pd_hourly_candles.close.isna())  & (pd_hourly_candles['datetime'].dt.time == pd.Timestamp('00:00:00').time())].iloc[0]['datetime']  # type: ignore
                 pd_hourly_candles = pd_hourly_candles[pd_hourly_candles.datetime>=first_candle_dt]
                 exchange_candles[symbol] = pd_hourly_candles
 
@@ -426,6 +435,45 @@ class YahooExchange:
 
         return exchange_candles
 
+def aggregate_candles(
+    interval : str,
+    pd_candles : pd.DataFrame
+) -> pd.DataFrame:
+    if interval[-1]=='m':
+        # 'm' for pandas means months!
+        interval = interval.replace('m','min')
+    pd_candles.set_index('datetime', inplace=True)
+    pd_candles_aggregated = pd_candles.resample(interval).agg({
+        'exchange' : 'first',
+        'symbol' : 'first',
+        'timestamp_ms' : 'first',
+        
+        'open': 'first',
+        'high': 'max',
+        'low': 'min',
+        'close': 'last',
+        'volume': 'sum',
+
+        'datetime_utc' : 'first',
+        'year' : 'first',
+        'month' : 'first',
+        'day' : 'first',
+        'hour' : 'first',
+        'minute' : 'first',
+        'dayofweek' : 'first',
+        'week_of_month' : 'first',
+
+        'apac_trading_hr' : 'first',
+        'emea_trading_hr' : 'first',
+        'amer_trading_hr' : 'first',
+
+        'pct_chg_on_close' : 'sum',
+
+    })
+    pd_candles.reset_index(inplace=True)
+    pd_candles_aggregated.reset_index(inplace=True)
+    return pd_candles_aggregated
+    
 def fetch_historical_price(
     exchange,
     normalized_symbol : str,
@@ -472,19 +520,21 @@ def fetch_candles(
     validation_max_gaps : int = 10,
     validation_max_end_date_intervals : int = 1
 ) -> Dict[str, Union[pd.DataFrame, None]]:
-    
+    exchange_candles = { '' : None }
+    num_intervals = int(candle_size.replace(candle_size[-1],''))
+
     if end_ts>datetime.now().timestamp():
         end_ts = int(datetime.now().timestamp())
 
     if type(exchange) is YahooExchange:
-        return exchange.fetch_candles(
+        exchange_candles = exchange.fetch_candles(
                             start_ts=start_ts,
                             end_ts=end_ts,
                             symbols=normalized_symbols,
                             candle_size=candle_size
                         )
     elif type(exchange) is NASDAQExchange:
-        return exchange.fetch_candles(
+        exchange_candles = exchange.fetch_candles(
                             start_ts=start_ts,
                             end_ts=end_ts,
                             symbols=normalized_symbols,
@@ -501,9 +551,9 @@ def fetch_candles(
             pd_candles = exchange_candles[symbol]
             if not pd_candles is None:
                 fix_column_types(pd_candles) # You don't want to do this from Futubull as you'd need import Futubull from there: Circular references
-        return exchange_candles
+                
     elif issubclass(exchange.__class__, CcxtExchange):
-        return _fetch_candles_ccxt(
+        exchange_candles = _fetch_candles_ccxt(
             start_ts=start_ts,
             end_ts=end_ts,
             exchange=exchange,
@@ -511,7 +561,41 @@ def fetch_candles(
             candle_size=candle_size,
             num_candles_limit=num_candles_limit
         )
-    return { '' : None }
+    if num_intervals!=1:
+        for symbol in exchange_candles:
+            if not exchange_candles[symbol] is None:
+                exchange_candles[symbol] = aggregate_candles(candle_size, exchange_candles[symbol]) #  type: ignore
+
+    # For invalid rows missing timestamps, o/h/l/c/v, fill forward close, set volume to zero.
+    for symbol in exchange_candles:
+        pd_candles = exchange_candles[symbol]
+        
+        if pd_candles is not None:
+            mask_invalid_candles = pd_candles["timestamp_ms"].isna()
+            if mask_invalid_candles.any():
+                pd_invalid_candles = pd_candles[mask_invalid_candles]
+
+                if logger is not None:
+                    logger.warning(f"Dropping {pd_invalid_candles.shape[0]}/{pd_candles.shape[0]} rows from {symbol} candles (null timestamp_ms)") # type: ignore
+                    logger.warning(f"{tabulate(pd_invalid_candles, headers='keys', tablefmt='psql')}") # type: ignore
+                    
+                def _to_timestamp_ms(dt):
+                    if pd.isna(dt):
+                        return pd.NA
+                    if isinstance(dt, str):
+                        dt = pd.to_datetime(dt)
+                    return int(dt.timestamp() * 1000)
+                
+                pd_candles.loc[mask_invalid_candles, "timestamp_ms"] = pd_candles.loc[mask_invalid_candles, "datetime"].apply(_to_timestamp_ms)
+
+                pd_candles["close"] = pd_candles["close"].ffill()
+                
+                pd_candles.loc[mask_invalid_candles, ["open", "high", "low"]] = pd_candles.loc[
+                                                                                    mask_invalid_candles, "close"
+                                                                                ].values[:, None] # type: ignore
+                pd_candles.loc[mask_invalid_candles, "volume"] = 0.0
+
+    return exchange_candles # type: ignore
 
 '''
 Find listing date https://gist.github.com/mr-easy/5185b1dcdd5f9f908ff196446f092e9b
@@ -563,7 +647,7 @@ def _fetch_candles_ccxt(
             
         def _calc_increment(candle_size):
             increment = 1
-            num_intervals = int(candle_size[0])
+            num_intervals = int(candle_size.replace(candle_size[-1],''))
             interval_type = candle_size[-1]
             if interval_type == "m":
                 increment = 60
